<?php

namespace Barn2\Plugin\Document_Library_Pro\Admin\Metabox;

use Barn2\DLP_Lib\Conditional,
	Barn2\DLP_Lib\Registerable,
	Barn2\DLP_Lib\Service,
	Barn2\DLP_Lib\Util,
	Barn2\Plugin\Document_Library_Pro\Util\SVG_Icon,
	Barn2\Plugin\Document_Library_Pro\Post_Type,
	Barn2\Plugin\Document_Library_Pro\Document;

defined( 'ABSPATH' ) || exit;

/**
 * Document Link - Edit Document Metabox
 *
 * @package   Barn2/document-library-pro
 * @author    Barn2 Plugins <info@barn2.com>
 * @license   GPL-3.0z
 */
class Document_Link implements Registerable, Service, Conditional {
	const ID = 'document_link';

	/**
	 * {@inheritdoc}
	 */
	public function is_required() {
		return Util::is_admin();
	}

	/**
	 * {@inheritdoc}
	 */
	public function register() {
		add_action( 'add_meta_boxes', [ $this, 'register_metabox' ], 1 );
		add_action( 'save_post_' . Post_Type::POST_TYPE_SLUG, [ $this, 'save' ] );
	}

	/**
	 * Register the metabox
	 */
	public function register_metabox() {
		add_meta_box(
			self::ID,
			__( 'Document Link', 'document-library-pro' ),
			[ $this, 'render' ],
			'dlp_document',
			'side',
			'high'
		);
	}

	/**
	 * Render the metabox.
	 *
	 * @param WP_Post $post
	 */
	public function render( $post ) {
		$document = new Document( $post->ID );

		$button_text         = $document->get_file_id() ? __( 'Replace File', 'document-library-pro' ) : __( 'Add File', 'document-library-pro' );
		$file_attached_class = $document->get_file_id() ? ' active' : '';
		$file_details_class  = $document->get_link_type() === 'file' ? 'active' : '';
		$url_details_class   = $document->get_link_type() === 'url' ? 'active' : '';

		?>
		<label for="<?php esc_attr( self::ID ); ?>" class="howto"><?php esc_html_e( 'Upload a file or add a URL where the document is located:', 'document-library-pro' ); ?></label>

		<!-- option selector -->
		<select name="_dlp_document_link_type" id="dlp_document_link_type" class="postbox">
			<option value="none" <?php selected( $document->get_link_type(), 'none' ); ?>><?php esc_html_e( 'None', 'document-library-pro' ); ?></option>
			<option value="file" <?php selected( $document->get_link_type(), 'file' ); ?>><?php esc_html_e( 'File Upload', 'document-library-pro' ); ?></option>
			<option value="url" <?php selected( $document->get_link_type(), 'url' ); ?>><?php esc_html_e( 'A custom URL', 'document-library-pro' ); ?></option>
		</select>

		<!-- file upload -->
		<div id="dlp_file_attachment_details" class="<?php echo esc_attr( $file_details_class ); ?>">
				<div class="dlp_file_attached <?php echo esc_attr( $file_attached_class ); ?>">
					<button type="button" id="dlp_remove_file_button">
						<span class="remove-file-icon" aria-hidden="true"></span>

						<span class="screen-reader-text">
						<?php
						/* translators: %s: File name */
						echo esc_html( sprintf( __( 'Remove file: %s', 'document-library-pro' ), $document->get_file_name() ) );
						?>
						</span>
					</button>

					<span class="dlp_file_name_text"><?php echo esc_html( $document->get_file_name() ); ?></span>
					<input id="dlp_file_name_input" type="hidden" name="_dlp_attached_file_name" value="<?php echo esc_attr( $document->get_file_name() ); ?>" />
				</div>


			<button id="dlp_add_file_button" class="button button-large"><?php echo esc_html( $button_text ); ?></button>
			<input id="dlp_file_id" type="hidden" name="_dlp_attached_file_id" value="<?php echo esc_attr( $document->get_file_id() ); ?>" />

		</div>

		<!-- direct url -->
		<div id="dlp_link_url_details" class="<?php echo esc_attr( $url_details_class ); ?>">
			<?php SVG_Icon::render( 'link' ); ?>
			<input type="text" id="dlp_direct_link_input" name="_dlp_direct_link_url" value="<?php echo esc_attr( $document->get_direct_link() ); ?>" placeholder="<?php echo esc_attr( 'https://' ); ?>" />
		</div>
		<?php
	}

	/**
	 * Save the metabox values
	 *
	 * @param mixed $post_id
	 */
	public function save( $post_id ) {
		// phpcs:ignore WordPress.Security.NonceVerification.Missing
		if ( ! isset( $_POST['_dlp_document_link_type'] ) ) {
			return;
		}

		$type = filter_input( INPUT_POST, '_dlp_document_link_type', FILTER_SANITIZE_FULL_SPECIAL_CHARS );
		$data = [];

		switch ( $type ) {
			case 'url':
				$data['direct_url'] = filter_input( INPUT_POST, '_dlp_direct_link_url', FILTER_SANITIZE_URL );
				break;

			case 'file':
				$data['file_id']   = filter_input( INPUT_POST, '_dlp_attached_file_id', FILTER_SANITIZE_NUMBER_INT );
				$data['file_name'] = filter_input( INPUT_POST, '_dlp_attached_file_name', FILTER_SANITIZE_FULL_SPECIAL_CHARS );
				break;
		}

		try {
			$document = new Document( $post_id );
			$document->set_document_link( $type, $data );
		// phpcs:ignore Generic.CodeAnalysis.EmptyStatement.DetectedCatch
		} catch ( \Exception $exception ) {
			// silent
		}
	}
}
