<?php

namespace Barn2\Plugin\Document_Library_Pro;

use Barn2\Plugin\Document_Library_Pro\Posts_Table_Pro\Table_Args,
	Barn2\Plugin\Document_Library_Pro\Posts_Table_Pro\Table_Query,
	Barn2\Plugin\Document_Library_Pro\Frontend_Scripts,
	Barn2\Plugin\Document_Library_Pro\Util\SVG_Icon,
	Barn2\Plugin\Document_Library_Pro\Taxonomies;

defined( 'ABSPATH' ) || exit;

/**
 * This class handles the folders HTML output
 *
 * @package   Barn2\document-library-pro
 * @author    Barn2 Plugins <support@barn2.com>
 * @license   GPL-3.0
 * @copyright Barn2 Media Ltd
 */
class Folder_Tree {
	private $exclude_categories = [];
	private $include_categories = [];
	private $exclude_terms      = [];
	private $include_terms      = [];

	/**
	 * Constructor.
	 *
	 * @param array $args
	 * @param array $atts
	 */
	public function __construct( $args = [], $atts = [] ) {
		if ( (int) $args['post_limit'] < 500 && (int) $args['post_limit'] !== -1 ) {
			$args['post_limit'] = 500;
		}

		$args['lazy_load'] = false;

		$this->args           = new Table_Args( $args );
		$this->query          = new Table_Query( $this->args );
		$this->shortcode_atts = $atts;

		$this->init_category_options();
	}

	/**
	 * Get any terms options associated to or excluded from the folder structure
	 */
	private function init_category_options() {
		$this->exclude_categories = array_filter( explode( ',', $this->args->get_args()['exclude_doc_category'] ) );
		$this->include_categories = array_filter( explode( ',', $this->args->get_args()['doc_category'] ) );

		$include_args        = $this->args->get_args()['term'];
		$this->include_terms = false !== strpos( $include_args, '+' ) ? array_filter( explode( '+', $include_args ) ) : array_filter( explode( ',', $include_args ) );

		$exclude_args        = $this->args->get_args()['exclude_term'];
		$this->exclude_terms = false !== strpos( $exclude_args, '+' ) ? array_filter( explode( '+', $exclude_args ) ) : array_filter( explode( ',', $exclude_args ) );
	}

	/**
	 * Retrieves the folder tree HTML
	 *
	 * @return  string  $folder_html
	 */
	public function get_html() {
		Frontend_Scripts::load_folder_scripts( $this->args );

		if ( ! $this->query->get_posts() ) {
			return __( 'No documents found.', 'document-library-pro' );
		}

		ob_start();
		$this->generate_category_folder_html( $this->generate_category_tree(), true, $this->shortcode_atts );

		return ob_get_clean();
	}

	/**
	 * Generates the full category tree.
	 *
	 * @return array
	 */
	private function generate_category_tree() {
		$categories = $this->get_attached_categories();
		$tree_root  = $this->generate_categories_tree_root( $categories );
		$tree       = $this->populate_missing_categories_into_tree( $tree_root );

		return $tree;
	}

	/**
	 * Retrieves all categories currently attached to the current table query
	 *
	 * @return  \WP_Term[]  $categories
	 */
	private function get_attached_categories() {
		$post_ids = array_map(
			function ( \WP_Post $post ) {
				return $post->ID;
			},
			$this->query->get_posts()
		);

		// We need to fetch all attached terms to match agains.
		// Query options provided to the shortcode could exclude certain categories (e.g. include, exclude, status etc..)

		$categories = wp_get_object_terms(
			$post_ids,
			[ Taxonomies::CATEGORY_SLUG ],
			[
				'orderby' => apply_filters( 'document_library_pro_folder_orderby', $this->args->get_args()['folders_order_by'] ),
				'order'   => $this->args->get_args()['folders_order']
			]
		);

		// Fetch empty parents not assigned to any documents in order to display the full hierarchy
		$category_parents_objects = [];

		foreach ( $categories as $category ) {
			$category_parents_ids = get_ancestors( $category->term_id, Taxonomies::CATEGORY_SLUG );

			if ( ! empty( $category_parents_ids ) ) {
				$term_objects = array_map(
					function( $parent_id ) {
						$parent_term = get_term( $parent_id, Taxonomies::CATEGORY_SLUG );

						return $parent_term;
					},
					$category_parents_ids
				);

				$category_parents_objects = array_merge( $category_parents_objects, $term_objects );
			}
		}

		$categories = array_merge( $categories, $category_parents_objects );

		// Adjust for shortcode provided category query changes
		if ( ! empty( $this->include_terms ) ) {
			$category_terms = array_map(
				function( $term ) {
					$term_split = explode( ':', $term, 2 );

					if ( $term_split[0] === 'doc_categories' ) {
						return $term_split[1];
					}
				},
				$this->include_terms
			);

			$using_term_ids = count( $category_terms ) === count( array_filter( $category_terms, 'is_numeric' ) );
			$field          = $using_term_ids ? 'id' : 'slug';

			$categories = array_filter(
				$categories,
				function( \WP_Term $item ) use ( $category_terms, $field ) {
					return in_array( $item->{$field}, $category_terms, true );
				}
			);
		}

		if ( ! empty( $this->exclude_terms ) ) {
			$category_terms = array_map(
				function( $term ) {
					$term_split = explode( ':', $term, 2 );

					if ( $term_split[0] === 'doc_categories' ) {
						return $term_split[1];
					}
				},
				$this->exclude_terms
			);

			$using_term_ids = count( $category_terms ) === count( array_filter( $category_terms, 'is_numeric' ) );
			$field          = $using_term_ids ? 'id' : 'slug';

			$categories = array_filter(
				$categories,
				function( \WP_Term $item ) use ( $category_terms, $field ) {
					return ! in_array( $item->{$field}, $category_terms, true );
				}
			);
		}

		if ( ! empty( $this->include_categories ) ) {
			$using_term_ids = count( $this->include_categories ) === count( array_filter( $this->include_categories, 'is_numeric' ) );
			$field          = $using_term_ids ? 'id' : 'slug';

			$categories = array_filter(
				$categories,
				function( \WP_Term $item ) use ( $field ) {
					return in_array( $item->{$field}, $this->include_categories, true );
				}
			);
		}

		if ( ! empty( $this->exclude_categories ) ) {
			$using_term_ids = count( $this->exclude_categories ) === count( array_filter( $this->exclude_categories, 'is_numeric' ) );
			$field          = $using_term_ids ? 'id' : 'slug';

			$categories = array_filter(
				$categories,
				function( \WP_Term $item ) use ( $field ) {
					return ! in_array( $item->{$field}, $this->exclude_categories, true );
				}
			);
		}

		return $categories;
	}


	/**
	 * Generate the top level of the tree array
	 *
	 * @param \WP_Term[] $categories
	 */
	private function generate_categories_tree_root( $categories ) {
		$category_tree = [];

		// Get all term ids
		$term_ids = array_map(
			function( $term ) {
				return $term->term_id;
			},
			$categories
		);

		// Check for those without included parents
		$top_level_categories = array_filter(
			$categories,
			function ( \WP_Term $category ) use ( $term_ids ) {
				return ! in_array( $category->parent, $term_ids, true );
			}
		);

		// Get top level term ids
		$top_level_ids = array_map(
			function ( $category ) {
				return $category->term_id;
			},
			$top_level_categories
		);

		$remaining_sub_categories = array_filter(
			$categories,
			function ( \WP_Term $category ) use ( $term_ids ) {
				return in_array( $category->parent, $term_ids, true );
			}
		);

		// Check if any of the top root categories are children of each other
		foreach ( $top_level_categories as $index => $category ) {
			if ( $category->parent === 0 ) {
				continue;
			}

			$ancestors = get_ancestors( $category->term_id, Taxonomies::CATEGORY_SLUG, 'taxonomy' );

			if ( ! empty( array_intersect( $ancestors, $top_level_ids ) ) ) {
				$remaining_sub_categories[] = $category;
				unset( $top_level_categories[ $index ] );
			}
		}

		// Get top level term ids so we can sort by menu order
		$top_level_ids = array_map(
			function ( $category ) {
				return $category->term_id;
			},
			$top_level_categories
		);

		$top_level_categories = get_terms(
			[
				'taxonomy'   => Taxonomies::CATEGORY_SLUG,
				'include'    => $top_level_ids,
				'orderby'    => apply_filters( 'document_library_pro_folder_orderby', $this->args->get_args()['folders_order_by'] ),
				'order'      => $this->args->get_args()['folders_order'],
				'hide_empty' => false,
			]
		);

		// Create the tree root (top level)
		foreach ( $top_level_categories as $category ) {
			$category_tree[ $category->term_id ] = $this->fill_in_subcategories( $remaining_sub_categories, $category->term_id );
		}

		return $category_tree;
	}

	/**
	 * Recursively fills in the attached subcategories of a tree root
	 *
	 * @param   \WP_Term[]   $categories
	 * @param   int         $parent
	 * @return  array       $category_tree
	 */
	private function fill_in_subcategories( $categories, $parent ) {
		$terms = array_filter(
			$categories,
			function( \WP_Term $category ) use ( $parent ) {
				return $category->parent === $parent;
			}
		);

		$category_tree = [];

		foreach ( $terms as $term ) {
			$category_tree[ $term->term_id ] = $this->fill_in_subcategories( $categories, $term->term_id );
		}

		return $category_tree;
	}

	/**
	 * Populates in missing category hierarchy of included terms
	 *
	 * @param   array   $category_tree
	 * @return  array   $category_tree
	 */
	private function populate_missing_categories_into_tree( $category_tree ) {
		foreach ( $category_tree as $category_id => $sub_categories ) {
			$term_children = get_terms(
				[
					'taxonomy'   => Taxonomies::CATEGORY_SLUG,
					'parent'     => $category_id,
					'orderby'    => apply_filters( 'document_library_pro_folder_orderby', $this->args->get_args()['folders_order_by'] ),
					'order'      => $this->args->get_args()['folders_order'],
					'hide_empty' => false,
				]
			);

			if ( ! empty( $this->exclude_categories ) ) {
				$using_term_ids = count( $this->exclude_categories ) === count( array_filter( $this->exclude_categories, 'is_numeric' ) );
				$field          = $using_term_ids ? 'id' : 'slug';

				$term_children = array_filter(
					$term_children,
					function( \WP_Term $item ) use ( $field ) {
						return ! in_array( $item->{$field}, $this->exclude_categories, true );
					}
				);
			}

			if ( ! empty( $this->exclude_terms ) ) {
				$category_terms = array_map(
					function( $term ) {
						$term_split = explode( ':', $term, 2 );

						if ( $term_split[0] === 'doc_categories' ) {
							return $term_split[1];
						}
					},
					$this->exclude_terms
				);

				$using_term_ids = count( $category_terms ) === count( array_filter( $category_terms, 'is_numeric' ) );
				$field          = $using_term_ids ? 'id' : 'slug';

				$term_children = array_filter(
					$term_children,
					function( \WP_Term $item ) use ( $category_terms, $field ) {
						return ! in_array( $item->{$field}, $category_terms, true );
					}
				);
			}

			foreach ( $term_children as $term ) {
				$term_descendants = get_term_children( $term->term_id, Taxonomies::CATEGORY_SLUG );

				if ( empty( $term_descendants ) && $term->count === 0 ) {
					continue;
				}

				$category_tree[ $category_id ][ $term->term_id ] = [];
			}

			$category_tree[ $category_id ] = $this->populate_missing_categories_into_tree( $category_tree[ $category_id ] );
		}

		return $category_tree;
	}

	/**
	 * Generates the folder HTML.
	 *
	 * @param mixed $categories
	 * @param bool $root
	 * @param array $shortcode_atts
	 */
	private function generate_category_folder_html( $categories, $root = false, $shortcode_atts = [] ) {
		?>
		<div class="dlp-folders-container">

			<?php
			if ( $root ) {
				$this->maybe_output_search_box( 'top' );
				$this->maybe_output_search_results_container();
			}
			?>

			<ul class="dlp-folders <?php echo $root ? 'dlp-folders-root' : ''; ?>" <?php echo $root ? 'data-shortcode-atts=\'' . wp_json_encode( $shortcode_atts ) . '\'' : ''; ?>">
				<?php
				foreach ( $categories as $category_id => $subcategories ) :
					$term = get_term( $category_id );
					?>
					<li class="dlp-folder dlp-folder closed" data-category-id="<?php echo esc_attr( $category_id ); ?>">
						<span class="dlp-category">
							<?php SVG_Icon::render( 'folder' ); ?>
							<?php SVG_Icon::render( 'folder_open' ); ?>
							<span class="dlp-folder-label">
								<span class="dlp-category-name"><?php echo esc_html( $term->name ); ?></span>
							</span>
						</span>

						<div class="dlp-folder-inner">
							<div class="dlp-category-table">
							</div>

							<div class="dlp-category-subcategories">
								<?php
								if ( ! empty( $subcategories ) ) {
									$this->generate_category_folder_html( $subcategories );
								}
								?>
							</div>
						</div>
					</li>
				<?php endforeach; ?>
			</ul>

			<?php
			if ( $root ) {
				$this->maybe_output_search_box( 'bottom' );
			}
			?>

		</div>
		<?php
	}

	/**
	 * Output the folder search box.
	 *
	 * @param string $context
	 */
	private function maybe_output_search_box( $context = 'top' ) {
		$search_position = $this->args->get_args()['search_box'];
		$search_html     = '';

		// Search Input
		if ( in_array( $search_position, [ $context, 'both' ], true ) ) {
			$search_query = filter_input( INPUT_GET, 'dlp_search', FILTER_SANITIZE_FULL_SPECIAL_CHARS ) ?? '';

			$search_html = sprintf(
				'<div class="dlp-folders-search"><label>%1$s<input type="search" value="%2$s"></label><div class="dlp-folders-reset"><a class="reset" href="#">%3$s %4$s</a></div></div>',
				esc_html__( 'Search:', 'document-library-pro' ),
				$search_query,
				SVG_Icon::get( 'reset' ),
				esc_html__( 'Reset', 'document-library-pro' )
			);
		}

		/* phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped */
		echo $search_html;
	}

	/**
	 * Output the search results container DIV.
	 */
	private function maybe_output_search_results_container() {
		$search_position = $this->args->get_args()['search_box'];

		if ( ! in_array( $search_position, [ 'top', 'bottom', 'both' ], true ) ) {
			return;
		}

		echo '<div class="dlp-folders-search-results"></div>';
	}
}
