<?php

namespace Barn2\Plugin\Document_Library_Pro;

use Barn2\Plugin\Document_Library_Pro\Util\Util,
	Barn2\DLP_Lib\Registerable,
	Barn2\DLP_Lib\Service;

defined( 'ABSPATH' ) || exit;

/**
 * Handle the DLP Search.
 *
 * @package   Barn2/document-library-pro
 * @author    Barn2 Plugins <info@barn2.com>
 * @license   GPL-3.0
 * @copyright Barn2 Media Ltd
 */
class Search_Handler implements Registerable, Service {

	/**
	 * {@inheritdoc}
	 */
	public function register() {
		add_action( 'init', [ $this, 'maybe_create_results_page' ] );
		add_action( 'template_redirect', [ $this, 'handle_search_redirect' ] );
		add_filter( 'the_content', [ $this, 'filter_search_page_content' ], 999, 1 );

		add_action( 'wp_head', [ $this, 'pre_wp_5_7_noindex' ] );
		add_filter( 'wp_robots', [ $this, 'wp_robots_noindex' ] );
	}

	/**
	 * Create the search results page if it doesn't exist.
	 */
	public function maybe_create_results_page() {
		if ( get_option( 'dlp_search_page' ) === false ) {
			Util::create_page(
				esc_sql( _x( 'document-search', 'Page slug', 'document-library-pro' ) ),
				'dlp_search_page',
				_x( 'Document Search Results', 'Page title', 'document-library-pro' ),
				'',
				0
			);
		}
	}

	/**
	 * Redirect to the search results page.
	 */
	public function handle_search_redirect() {
		if ( empty( $_GET['dlp_search'] ) ) {
			return;
		}

		$search_results_page_id = get_option( 'dlp_search_page' );

		if ( $search_results_page_id && ! is_page( (int) $search_results_page_id ) ) {
			wp_safe_redirect( add_query_arg( 'dlp_search', $_GET['dlp_search'], get_permalink( $search_results_page_id ) ), 301 );
			exit;
		}
	}

	/**
	 * Filter the_content to display search results.
	 *
	 * @param string $content
	 * @return string
	 */
	public function filter_search_page_content( $content ) {

		$search_results_page_id = get_option( 'dlp_search_page' );
		$search_query           = filter_input( INPUT_GET, 'dlp_search', FILTER_SANITIZE_FULL_SPECIAL_CHARS );

		if ( ! $search_results_page_id || ! is_page( (int) $search_results_page_id ) || ! in_the_loop() || ! is_main_query() ) {
			return $content;
		}

		remove_filter( 'the_content', [ $this, 'filter_search_page_content' ], 999 );

		if ( empty( $search_query ) ) {
			return sprintf( '<h4>%s</h4>', esc_html__( 'No matching documents', 'document-library-pro' ) );
		}

		return dlp_get_doc_library(
			[
				'folders'     => false,
				'search_term' => $search_query
			]
		);
	}

	/**
	 * Noindex the search page. (< WP 5.7)
	 */
	public function pre_wp_5_7_noindex() {
		// wp_no_robots is deprecated since WP 5.7.
		if ( function_exists( 'wp_robots_no_robots' ) ) {
			return;
		}

		if ( is_page( get_option( 'dlp_search_page' ) ) ) {
			wp_no_robots();
		}
	}


	/**
	 * Noindex the search page. (>= WP 5.7)
	 *
	 * @param array $robots
	 * @return array
	 */
	public function wp_robots_noindex( array $robots ) {
		if ( is_page( get_option( 'dlp_search_page' ) ) ) {
			return wp_robots_no_robots( $robots );
		}

		return $robots;
	}
}
